import os
from src.param import command
from . import utility


def make(sources: list[str], output: str, *tools) -> list[str]:
    if len(tools) == 0:
        raise Exception("")
    if len(sources) == 0:
        raise Exception("")

    cmds: list[command.Command] = list()
    temp_file: list[str] = []

    if hasattr(tools[0], "requires_extent_name"):
        tools = (utility.Copy(),) + tools

    current_sources = sources
    for first, second in zip(tools, tools[1:]):
        extent_name = "" if not hasattr(second, "requires_extent_name") else second.requires_extent_name()
        os.makedirs(os.path.dirname(os.path.abspath(output)), exist_ok=True)
        step_output = os.path.join(os.path.dirname(output), os.path.basename(current_sources[
                                                                                 0]) + extent_name)  # os.path.splitext(os.path.basename(current_sources[0]))[0] + extent_name)
        first.set_sources(current_sources)
        first.set_output(step_output)
        current_sources = [step_output]
        cmds.append(first.create_cmd())
        temp_file.append(step_output)

    if len(tools) > 1:
        tools[-1].set_sources(current_sources)
        tools[-1].set_output(output)
        cmds.append(tools[-1].create_cmd())

    command.run(cmds)
    return temp_file



from . import param


class MultiValue(param.Param):

    def __init__(self, prefix: str = "", values: list[str] = None):
        self.prefix = prefix
        self.values = values or []

    def is_active(self) -> bool:
        return len(self.values) > 0

    def get_value(self) -> str:
        return " ".join(f"{self.prefix}{value}" for value in self.values)



from . import param


class Optional(param.Param):
    def __init__(self, value: str, active: bool = False) -> None:
        self.value = value
        super().__init__(active)

    def get_value(self) -> str:
        return self.value



from src.param import command
from src.param.optional import Optional
from src.param.single_value import SingleValue
from src.param.multi_value import MultiValue
from src.param.make import make_param_list_, make_class_param_list


class GccInfo:
    def __init__(self,
                 include_dirs: list[str] = None,
                 is_static: bool = False,
                 md: bool = False,
                 debugger: bool = False,
                 optimize: str = None,
                 flag: list[str] = None,
                 nos: list[str] = None,
                 ):
        self.include_dirs: list[str] = include_dirs
        self.is_static: bool = is_static
        self.md: bool = md
        self.debugger: bool = debugger
        self.optimize: str = optimize
        self.flag: list[str] = flag
        self.nos: list[str] = nos


class Gcc:
    def __init__(self,
                 target: str = "g++",
                 info: GccInfo = None,
                 sources: list[str] = None,
                 output: str = None,
                 ):
        self.target: str = target

        info = info or GccInfo()

        self._sources: MultiValue = MultiValue(values=sources)
        self._output: SingleValue = SingleValue("-o", output, output is not None)
        self._include_dirs: MultiValue = MultiValue(prefix="-I", values=info.include_dirs)
        self._is_static: Optional = Optional("-static", info.is_static)
        self._md: Optional = Optional("-MD", info.md)
        self._debugger: Optional = Optional("-g", info.debugger)
        self._optimize: SingleValue = SingleValue("-O", info.optimize or "", info.optimize is not None)
        self._flags: dict[str, Optional] = {}
        self._nos: dict[str, Optional] = {}

        flag_no = [
            "builtin", "strict-aliasing", "omit-frame-pointers", "stack-protector", "pie", "pic", "exceptions", "rtti"
        ]
        flag = ["permissive"]
        no = ["stdlib"]
        for f in flag_no:
            self._flags[f"no-{f}"] = Optional(f"-fno-{f}")
        for f in flag:
            self._flags[f] = Optional(f"-f{f}")
        for n in no:
            self._nos[n] = Optional(f"-no{n}")

        if info.flag:
            for f in info.flag:
                if f not in self._flags:
                    continue
                self._flags[f].on()

        if info.nos:
            for n in info.nos:
                if n not in self._nos:
                    continue
                self._nos[n].on()

    def requires_extent_name(self) -> str:
        return ".cpp"

    def create_cmd(self) -> command.Command:
        result: command.Command = command.Command(self.target)
        result.option += make_class_param_list(self)
        result.option += make_param_list_([f for f in self._flags.values()])
        result.option += make_param_list_([n for n in self._nos.values()])
        return result

    def set_sources(self, sources: list[str]) -> None:
        self._sources.values = sources or []

    def set_output(self, output: str) -> None:
        self._output.set_value(output)
        self._output.on()


class CppInfo(GccInfo):
    pass

class Cpp(Gcc):
    def __init__(self,
                 target: str = "gcc",
                 info: CppInfo = None,
                 sources: list[str] = None,
                 output: str = None,
                 ):
        super().__init__(target, info, sources, output)

    def create_cmd(self) -> command.Command:
        result: command.Command = super().create_cmd()
        result.option.insert(0, "-E")
        return result

    def requires_extent_name(self) -> str:
        return ".c"



from . import param


def get_class_param_list(obj) -> list[param.Param]:
    variable_ids: list[str] = [attr for attr in dir(obj) if
                               not callable(getattr(obj, attr)) and not attr.startswith("__") and isinstance(getattr(obj,attr),
                                                                                                             param.Param)]
    return [getattr(obj, variable) for variable in variable_ids]


def make_param_list(*options) -> list[str]:
    return make_param_list(list[param.Param](options))


def make_param_list_(options: list[param.Param]) -> list[str]:
    result: list[str] = list()
    for option in options:
        if not option.is_active():
            continue
        result.append(option.get_value())
    return result


def make_class_param_list(obj) -> list[str]:
    return make_param_list_(get_class_param_list(obj))



from src.param import command


class Copy:

    def __init__(self):
        self.output = None
        self.sources = None

    def set_sources(self, sources: list[str]) -> None:
        self.sources = sources

    def set_output(self, output: str) -> None:
        self.output = output

    def create_cmd(self) -> command.Command:
        result = command.Command("cp")
        result.option += self.sources
        result.option.append(self.output)
        return result



import subprocess


class Command:

    def __init__(self, target: str, option=None):
        self.target: str = target
        self.option: list[str] = option or []

    def as_list(self) -> list[str]:
        return [self.target] + self.option

    def run(self) -> None:
        print(self.as_list())
        subprocess.run(self.as_list())


def run(*cmds):
    if len(cmds) == 1 and isinstance(cmds[0], list):
        for cmd in cmds[0]:
            cmd.run()
    else:
        for cmd in cmds:
            cmd.run()



from abc import ABC, abstractmethod


class Param(ABC):

    def __init__(self, active: bool = False) -> None:
        self._active: bool = active

    def is_active(self) -> bool:
        return self._active

    def set_active(self, value: bool) -> None:
        self._active = value

    def on(self):
        self.set_active(True)

    def off(self) -> None:
        self.set_active(False)

    @abstractmethod
    def get_value(self) -> str:
        pass



import setuptools
from setuptools import setup

setup(
    name='namke',
    version='1.0.0',
    description='nagisa make, for nasm, c and c plus plus',
    author='nagisa',
    author_email='1300296933@qq.com',
    package_dir={"": "src"},  # 打包目录
    packages=setuptools.find_packages(where='src'),  # 搜索存在__init__.py打包文件夹
)



from . import param


class SingleValue(param.Param):

    def __init__(self, prefix: str, value: str = "", active: bool = False):
        self.prefix = prefix
        self.value = value
        super().__init__(active)

    def set_value(self, new_value: str):
        self.value = new_value

    def get_value(self) -> str:
        return self.prefix + self.value




from src.param import command
from src.param.optional import Optional
from src.param.multi_value import MultiValue
from src.param.single_value import SingleValue
from src.param.make import make_class_param_list

class NasmInfo:
    def __init__(self,
                 is_static: bool = False,
                 md: bool = False,
                 debugger: str = None,
                 format: str = None,
                 ):
        self.is_static: bool = is_static
        self.md: bool = md
        self.debugger: str = debugger
        self.format: str = format

class Nasm:

    def __init__(self,
                 target: str = "nasm",
                 info: NasmInfo = None,
                 sources: list[str] = None,
                 output: str = None,
                 ):
        self.target: str = target

        info = info or NasmInfo()

        self._sources: MultiValue = MultiValue(values=sources)
        self._output: SingleValue = SingleValue(prefix="-o", value=output, active=output is not None)
        self._is_static: Optional = Optional("-static", info.is_static)
        self._1_md: Optional = Optional("-MD", info.md)
        self._2_debugger: SingleValue = SingleValue("-g", info.debugger, info.debugger is not None)
        self._format: SingleValue = SingleValue("-f ", info.format, info.format is not None)

    def create_cmd(self) -> command.Command:
        return command.Command(self.target, make_class_param_list(self))

    def set_sources(self, sources: list[str]) -> None:
        self._sources.values = sources or []

    def set_output(self, output: str) -> None:
        self._output.set_value(output)
        self._output.on()

    def requires_extent_name(self) -> str:
        return ".asm"